#include <iostream>
#include <windows.h>
#define DEBUG 0

using namespace std;
void printHelp(void);
void printHeader(void);
void printPrompt(void);
char getCommand(void);
int procCommand(char,char*&,int*);
int timedCapture(char*&,char);
int printBuffer(char*&,int);
int saveBuffer(char*&,int);
int clearBuffer(char*&);
char setCommPort(void);

int main(int argc, char* argv){
	
	char cmd('x');
	char* buffer = NULL;
	int samples=0;
	printHeader();
	printHelp();
	
	while(cmd != 'q'){
	printPrompt();
	cmd = getCommand();
	procCommand(cmd,buffer,&samples);
	}
	
	cout << "\nGoodbye!\n\n";
	
	
}

// Function to capture from serial port for some number of seconds.
int timedCapture(char*& buffer,char commPort){
	// Local variables.
	char prompt[5]; // Local buffer to see if the prompt is valid #R$ or non-valid.
	int time=0;
	char* commString = new char(20); // To hold the COM Port name for init.
	int samples=0;
	DCB dcb; // windows.h structure for setting up the COM port.
	
	// If the buffer isn't null, we should release the memory stored there before assigning it to a new location.
	if (buffer != NULL) clearBuffer(buffer);
	
	// See how long we should capture for, and how many samples we need to take.
	cout<<"\nTimed capture - Enter Time (s)\t >  ";
	cin >> time;
	samples = 250 * time + 1;
	int needBytes = samples*9 ; // 9 bytes are needed for each sample for ADXL350RS serial dongle.
	
	// Initialize the new capture buffer.
	cout <<"\nInitializing capture buffer ...";
	buffer = new char[needBytes+5]; // + 5 for headers and prompts and breathing room.
	
	cout<<"\nInitializing serial port ... ";
	cout<<"\nPreparing COM" << commPort << " to capture for "<<time<<" seconds at 250 Hz";
	
	// Initialize overlapped structure.
	// The overlapped structure is required for win32 api COM port communications in asynch mode.
	HANDLE status = CreateEvent(NULL,FALSE,FALSE,NULL); // required event handler for overlapped structure to manipulate.
	OVERLAPPED overlap; // the overlapped structure
	overlap.Offset = 0;
	overlap.OffsetHigh = 0;
	overlap.hEvent = status;


	// Set up the COM port name initialization structure.
	strcpy(commString,"\\\\.\\COM"); // this is how windows identifies the com ports (strangely)
	int len = strlen(commString);
	if (DEBUG) cout <<"\nString Length: "<<len;
	commString[7] = commPort; // Append the COM port number to the init string.
	commString[8] = '\0';	// make sure it's null terminated.
	
	// Open the COM port handle.
	HANDLE hCOM=CreateFile(commString,GENERIC_READ | GENERIC_WRITE,0,NULL,OPEN_EXISTING,FILE_FLAG_OVERLAPPED,NULL); 
	
	// Populate the DCB structure with current data, then modify to work with SERIAL ACCELEROMETER.
	if (GetCommState(hCOM,&dcb) == 0){
		cout <<"\nERROR READING COM PORT DCB";
	}
	dcb.BaudRate = 115200;
	dcb.fParity = 0;
	dcb.fOutX = FALSE;
	dcb.fInX = FALSE;
	dcb.ByteSize = 8;
	dcb.StopBits = ONESTOPBIT;
	
	
	// Implement our changes to the COM port handle instance.
	if (SetCommState(hCOM,&dcb) == 0){
			cout<< "\nERROR: " << GetLastError();
	}
	
	// Purge out any data that was in the COM port buffer before we got there.
	if (PurgeComm(hCOM,PURGE_RXCLEAR | PURGE_TXCLEAR) == 0){ // purge serial IO buffers.  Returns 0 on failure.
		cout<<"\n\tERROR PURGING COM PORT";
	}
	
	// If hCOM opened okay, read in the data to our data buffer.
	if (hCOM != INVALID_HANDLE_VALUE){
		// continue with reading.
		char towrite = '?'; // holds the command we send to SERIAL COM DEVICE.
		char trash;
		DWORD bytes_read = 0;
		DWORD bytes_written = 0;
		
		// Check for the ready prompt to ensure data is available.
		if (ReadFile(hCOM,prompt,samples,&bytes_read,&overlap)==0){
			if (GetOverlappedResult(hCOM,&overlap,&bytes_read,TRUE) == 0){
				cout<<"\nERROR reading COM Port!";
				cout<< "\nERROR: " << GetLastError();
			}
			else cout<<"\nChecking for SERIAL DONGLE in BINARY MODE: ";
		}
		
		if (prompt[0] == '#' && prompt[1] == 'R' && prompt[2] =='$'){
			cout<<" OK";
			cout<<"\nChecks complete.\n\nPress ENTER to begin capturing data.";
			fflush(stdin);
			cin.get(); // Wait on user to initiate data capture.
		
		// Purge the buffer after the user wants to start.	
		if (PurgeComm(hCOM,PURGE_RXCLEAR | PURGE_TXCLEAR) == 0){ // purge serial IO buffers.  Returns 0 on failure.
			cout<<"\n\tERROR PURGING COM PORT";
		}
		
		cout << "\nSending start signal ...";
		towrite = 'S'; // send the start character 0x53
		if (WriteFile(hCOM,&towrite,1,&bytes_written,&overlap)==0) {
			if (GetOverlappedResult(hCOM,&overlap,&bytes_written,TRUE) == 0){
			cout<<"\nERROR writing COM port!";
			cout<< "\nERROR: " << GetLastError();
			}
		}
		cout <<"\nReading data, please wait for "<<time <<" seconds.";
		
		// Read into data buffer, 9 bytes per sample.
		if (ReadFile(hCOM,buffer,needBytes,&bytes_read,&overlap)==0){
			if (GetOverlappedResult(hCOM,&overlap,&bytes_read,TRUE) == 0){
				cout<<"\nERROR reading COM Port!";
				cout<< "\nERROR: " << GetLastError();
			}
			else cout<<"\nBytes read: "<<bytes_read;
		}
			
			
			
		}
		
		// If the ready prompt didn't show up, the device has a problem.
		else {
			cout <<"\nERROR:  DEVICE NOT FOUND.  VERIFY DEVICE IS ATTACHED AND SET TO BINARY MODE";
		}
		
		// Close the serial port.
		CloseHandle(hCOM);
		
		// Function returns the number of samples it read.
		return samples;
		
	}
	
	// If we couldn't open the com port, whine about it some.
	else cout<<"\nERROR OPENING COM PORT!";
	
}

// This function works same as saveBuffer function.  See comments there.
int printBuffer(char*& buffer,int samples){
	
	int bufferBytes = 0;
	bufferBytes = samples*9 + 5; // 9 bytes per sample.
	char start_char,data_char,end_char;
	WORD x_axis=0,y_axis=0,sample=0;
	WORD bL=0,bH=0,xH=0,xL=0,yH=0,yL=0;
	float x_gval=0,y_gval=0;
	
	float time;
	
	if (buffer == NULL){
		cout <<"\nERROR: Data buffer has no data.";
		return 1;
	}
	
	int count = 1;
	for (int i=3; i<bufferBytes-9; i+=9){
		bH=0,bL=0;
		xH=0,xL=0;
		yH=0,yL=0;
		x_gval=0,y_gval=0;
		
		if (&buffer[i] == NULL) cout<<"\nNULL POINTER IN LOOP!";
		start_char = buffer[i+0];
		data_char = buffer[i+1];
		memcpy(&bH,&buffer[i+2],1);
		memcpy(&bL,&buffer[i+3],1);
		memcpy(&xH,&buffer[i+4],1);
		memcpy(&xL,&buffer[i+5],1);
		memcpy(&yH,&buffer[i+6],1);
		memcpy(&yL,&buffer[i+7],1);
		end_char = buffer[i+8];
		sample = (bH << 8)| bL;
		x_axis = (xH << 8) | xL;
		y_axis = (yH << 8) | yL;
		time = (float)sample / 250;
	
		if (start_char != '#' || data_char != '@' || end_char != '$' ){
			return 0;
		}
		x_gval = (float)(x_axis - 517) / 20.5;
		y_gval = (float)(y_axis - 518) / 20.5;

		printf("\n%f,%f,%f",time,x_gval,y_gval);
		
	}
		
	return 0;
}


int saveBuffer(char*& buffer,int samples){
	
	// Local variables
	char* filename = new char(255);
	FILE* fp; // file handle for saving to disk.
	int bufferBytes = 0;
	
	// Data for interpreting the 9-byte output string.
	char start_char,data_char,end_char;
	WORD x_axis=0,y_axis=0,sample=0;
	WORD bL=0,bH=0,xH=0,xL=0,yH=0,yL=0;
	float x_gval=0,y_gval=0;	
	float time;	
	
	
	
	// Check that a data buffer exists for saving.
	if (buffer == NULL){
		cout <<"\nERROR: Data buffer has no data.";
		return 1;
	}
	
	// Get the file name to save the data buffer to.
	cout <<"\n\nSave to file name : > ";
	cin >> filename;
	cout <<"\nSaving: "<<filename;
	
	// Open the file for writing - does not check for existing files.
	fp = fopen(filename,"w");
	if (fp == NULL ) {
		cout <<"\nERROR:  Failed to open file \"" << filename <<"\"";
		cout <<"\nAborting file write";
		return 1;
	}
	
	
	// Figures how many bytes are to be written, based on the number of samples acquired.
	bufferBytes = samples*9 + 5; // 9 bytes per sample.
	
	// Write the data in buffer to a file.
	int count = 1;
	for (int i=3; i<bufferBytes-9; i+=9){
		bH=0,bL=0;
		xH=0,xL=0;
		yH=0,yL=0;
		x_gval=0,y_gval=0;
			
		// Check for NULL pointers / buffer overruns.
		if (&buffer[i] == NULL) cout<<"\nNULL POINTER IN LOOP!";
			
		// Read the 9-byte data structure into local variables
		start_char = buffer[i+0];
		data_char = buffer[i+1];
		memcpy(&bH,&buffer[i+2],1);
		memcpy(&bL,&buffer[i+3],1);
		memcpy(&xH,&buffer[i+4],1);
		memcpy(&xL,&buffer[i+5],1);
		memcpy(&yH,&buffer[i+6],1);
		memcpy(&yL,&buffer[i+7],1);
		end_char = buffer[i+8];
		
		// Conmbine the high and low bytes to make meaningful data.
		sample = (bH << 8)| bL;
		x_axis = (xH << 8) | xL;
		y_axis = (yH << 8) | yL;

		// Convert into seconds, based on 250Hz sampling rate.
		time = (float)sample / 250;
	
		// If we don't get a valid data structure, return.  Not sure why this happens, but it does.
		if (start_char != '#' || data_char != '@' || end_char != '$' ){
			break;
		}
		
		// Convert raw data into meaningful G force data.  This is device specific.  
		// The values below work well for mine.
		x_gval = (float)(x_axis - 517) / 20.5;
		y_gval = (float)(y_axis - 518) / 20.5;

		// Print the line to a file
		fprintf(fp,"\n%f,%f,%f",time,x_gval,y_gval);
	}
	
	// Close the file.
	fclose(fp);
	cout <<"\nFile saved.\n\n";
	return 0;
}

// Function to free up the memory in the data buffer.
int clearBuffer(char*& buffer){
	cout << "Clearing buffer ... ";
	delete [] buffer;
	buffer = NULL;
	cout <<" OK.";
	return 0;
}

// Function dispatches commands to other functions, controls program flow.
int procCommand(char cmd,char*& buffer,int* num_samples){
	// Initialize commPort, defaults to COM1.
	static char commPort='1';
	
	// Dispatch commands
	if (cmd == 't') *num_samples = timedCapture(buffer,commPort);
	if (cmd == 'p') printBuffer(buffer,*num_samples);
	if (cmd == 's') saveBuffer(buffer,*num_samples);
	if (cmd == 'c') clearBuffer(buffer);
	if (cmd == 'h') printHelp();
	if (cmd == 'z') {
		cout <<"\nCOM Port is currently: COM"<<commPort;
		commPort = setCommPort();
	}
	
	
	// Check to see if command is unknown, and offer help if it is.
	if (cmd != 't' && cmd != 'p' && cmd != 's' && cmd != 'c' && cmd != 'z' && cmd != 'q' && cmd !='h') 
		cout<<"\nUnknown command.  Enter 'h' for help.";
}

// Function to get command from user
char getCommand(void){
	char cmd;
	cin >> cmd;
	return cmd;
}


// Function to print the prompt screen.
void printPrompt(void){
	cout<<"\nSerAccel ?>";
}


// Function to print the header when opening program.
void printHeader(void){
	cout<<endl;
	cout<<"############################################"<<endl;
	cout<<"#                                          #"<<endl;
	cout<<"#  Serial Data Logger R1.1   -  D. Yates   #"<<endl;
	cout<<"#                                          #"<<endl;
	cout<<"############################################"<<endl;
}

// Prints a summary of the commands available.
void printHelp(void){
	cout<<"\n\tt - Timed capture\
			\n\tp - Print buffer\
			\n\tz - Set COM Port\
			\n\ts - Save buffer to disk\
			\n\tc - Clear data buffer\
			\n\th - Print help\
			\n\tq - Quit\
			\n\t";
}

// Gets a COM port character to append to the "COM" string.  Used for configuring with custom COM ports.
char setCommPort(void){
	char comm;
	cout <<"\nEnter COM Port number: > ";
	cin >> comm;
	cout <<"Setting COM Port to COM"<<comm;
	return comm;
}

